﻿using Microsoft.Extensions.ObjectPool;
using OpenCvSharp;
using System;
using System.Collections.Generic;
using System.Net.Sockets;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;
using System.Threading.Tasks.Dataflow;
using WindowsFormsApp1.FFmpeg;

namespace WindowsFormsApp1
{
    internal class FpsWriter : IDisposable
    {
        // 윈도우 운영체제의 시간 함수 (1) - GetTickCount와 timeGetTime의 차이점
        // ; https://www.sysnet.pe.kr/2/0/11063
        [DllImport("winmm.dll", EntryPoint = "timeBeginPeriod")]
        public static extern uint timeBeginPeriod(uint uMilliseconds);

        [DllImport("winmm.dll", EntryPoint = "timeEndPeriod")]
        public static extern uint timeEndPeriod(uint uMilliseconds);

        [DllImport("winmm.dll", EntryPoint = "timeGetTime")]
        public static extern uint timeGetTime();

        readonly int _maxQueueLength;

        // C# - DefaultObjectPool<T>의 IDisposable 개체에 대한 풀링 문제
        // ; https://www.sysnet.pe.kr/2/0/12894
        ObjectPool<Mat> _pool;

        // C# - BufferBlock<T> 사용 예제
        // ; https://www.sysnet.pe.kr/2/0/12845
        readonly BufferBlock<FrameItem> _queue;

        readonly Task _frameTask;
        readonly FFmpegVideoWriter _writer;
        readonly EventWaitHandle _processTaskExit;

        FrameItem _lastCaptured = new FrameItem();

        internal struct FrameItem
        {
            public uint TimeStamp;
            public Mat Mat;

            public override string ToString()
            {
                return $"Frame at {TimeStamp}";
            }
        }

        public FpsWriter(FFmpegVideoWriter writer, MatType matType)
        {
            timeBeginPeriod(1);

            _writer = writer;
            _maxQueueLength = (int)writer.Fps * 3; // 3초 분량

            IPooledObjectPolicy<Mat> policy = new PooledObjectPolicy(_writer.FrameSize, MatType.CV_8UC3);

            var provider = new DefaultObjectPoolProvider();
            provider.MaximumRetained = _maxQueueLength;
            _pool = provider.Create(policy);

            _queue = new BufferBlock<FrameItem>();
            _processTaskExit = new EventWaitHandle(false, EventResetMode.AutoReset);
            _frameTask = Task.Run(async () => await ProcessFrame());
        }

        private async Task ProcessFrame()
        {
            try
            {
                uint _started = 0;
                Queue<FrameItem> frameQueue = new Queue<FrameItem>();
                FrameItem lastFrameOfPreviousTime = new FrameItem();
                uint previousFrameTimeStamp = 0;

                while (true)
                {
                    FrameItem frame = await _queue.ReceiveAsync();
                    if (frame.TimeStamp == 0)
                    {
                        break;
                    }

                    // 키가 중복되는 것을 방지 (1밀리 초 내에 화면 캡처 데이터가 오는 것?)
                    if (previousFrameTimeStamp == frame.TimeStamp)
                    {
                        ReleaseMat(frame);
                        continue;
                    }

                    previousFrameTimeStamp = frame.TimeStamp;

                    AddFrame(frameQueue, ref lastFrameOfPreviousTime, frame, ref _started);
                }

                // frameQueue에 남아 있는 이미지들을 기록
                uint remainedTime = previousFrameTimeStamp - _started;
                WriteRemainedImages(frameQueue, remainedTime, lastFrameOfPreviousTime, _started);
            }
            catch (Exception e)
            {
                System.Diagnostics.Trace.WriteLine(e.ToString());
            }

            _processTaskExit.Set();
        }

        private void WriteRemainedImages(Queue<FrameItem> frameQueue, uint remainedTime, FrameItem lastFrameOfPreviousTime, uint fromTime)
        {
            MakeFramesPerTime(frameQueue, ref lastFrameOfPreviousTime, fromTime, 1, remainedTime + 1);
            ReleaseMat(lastFrameOfPreviousTime);
        }

        private void ReleaseMat(FrameItem frame)
        {
            if (frame.Mat != null)
            {
                _pool.Return(frame.Mat);
                frame.Mat = null;
            }
        }

        internal void AddFrame(Queue<FrameItem> frameQueue, ref FrameItem lastFrameOfPreviousTime,
            FrameItem frame, ref uint started)
        {
            frameQueue.Enqueue(frame);

            if (started == 0) // 첫 번째 프레임인 경우
            {
                started = frame.TimeStamp;
                return;
            }

            uint diff = frame.TimeStamp - started;
            if (diff >= 1000) // 쌓인 프레임의 처음과 마지막의 프레임 시간 차이가 1초 넘으면, Queue에 쌓여 있던 프레임을 파일에 fps에 맞게 기록 
            {
                uint seconds = (uint)(diff / 1000);
                MakeFramesPerTime(frameQueue, ref lastFrameOfPreviousTime, started, seconds, 1000);
                started += (seconds * 1000);
            }
        }

        internal void MakeFramesPerTime(Queue<FrameItem> savedFrames, ref FrameItem lastFrameOfPreviousTime,
            uint fromTime, uint gapCount, uint gapTime)
        {
            List<FrameItem> framesDuringTime = new List<FrameItem>();
            FrameItem lastFrame = lastFrameOfPreviousTime;

            uint totalFrames = (uint)(_writer.Fps * (gapTime / 1000.0));

            for (uint time = 0; time < gapCount; time ++)
            {
                uint endTime = fromTime + ((time + 1) * gapTime);

                // 기준 시점 전의 시간 기록을 갖는 frame이 있다면 보정용으로 그것도 추가
                // 예를 들어, 999에 캡처된 프레임과 1100에 캡처된 프레임이 있는 경우,
                // 1001 시점을 표현할 프레임에 대해 1100보다는 999에 캡처된 프레임을 사용하기 위해.
                framesDuringTime.Add(lastFrame);

                FrameItem frame = new FrameItem();

                // gapTime 초 내의 시간 기록을 갖는 frame을 가져옴
                while (true)
                {
                    if (savedFrames.TryPeek(out frame) == false)
                    {
                        break;
                    }

                    if (frame.TimeStamp >= endTime)
                    {
                        break;
                    }

                    lastFrame = frame;

                    savedFrames.Dequeue();
                    framesDuringTime.Add(frame);
                }

                // 1초 후의 시간 기록을 갖는 frame이 있다면 보정용으로 그것도 추가
                // 예를 들어, 960에 캡처된 프레임과 1001에 캡처된 프레임이 있는 경우,
                // 997 시점을 표현할 프레임에 대해 960보다는 1001에 캡처된 프레임을 사용하기 위해.

                FrameItem firstFrameOfNextTime = new FrameItem();
                savedFrames.TryPeek(out firstFrameOfNextTime);
                framesDuringTime.Add(firstFrameOfNextTime);

                // _writer.Fps == 30인 경우,
                // 33ms마다 1개의 frame을 써서 총 30개의 이미지를 기록
                WriteFrames(framesDuringTime, fromTime, this.FrameGapTimePerSecond, totalFrames);
                
                // 아래의 과정에서 (...캡처 이미지...)에 해당하는 데이터만 메모리 해제
                // [이전 시간의 마지막 프레임, (...캡처 이미지...), 마지막 캡처 이미지, 다음 시간의 첫 번째 프레임]
                for (int i = 1; i < framesDuringTime.Count - 2; i++)
                {
                    FrameItem item = framesDuringTime[i];
                    ReleaseMat(item);
                }

                if (lastFrame.TimeStamp != lastFrameOfPreviousTime.TimeStamp)
                {
                    ReleaseMat(lastFrameOfPreviousTime);
                    lastFrameOfPreviousTime = lastFrame;
                }
            }
        }

        internal void TrySleep(int elapsed)
        {
            int delay = (int)this.FrameGapTimePerSecond / 2;
            if (elapsed > delay)
            {
                return;
            }

            Thread.Sleep(delay);
        }

        uint FrameGapTimePerSecond
        {
            get { return (uint)(1000 / _writer.Fps); }
        }

        internal void WriteFrames(List<FrameItem> framesDuringTime, uint fromTime, uint frameGapTimePerSecond, uint totalFrameCount)
        {
            if (framesDuringTime.Count == 0)
            {
                return;
            }

            // fps에 해당하는 시간의 프레임을 VideoWriter에 기록
            // 일례로, 30fps의 경우 33ms마다 1개의 프레임이 있어야 함.
            uint frameNo = 0;
            int startPos = 0;

            while (frameNo < totalFrameCount)
            {
                frameNo++;

                uint frameTime = fromTime + (frameGapTimePerSecond * (frameNo - 1));
                startPos = FindNearestFrame(framesDuringTime, startPos, frameTime);

                try
                {
                    _writer.Write(framesDuringTime[startPos].Mat);
                }
                catch (Exception ex)
                {
                    System.Diagnostics.Trace.WriteLine(ex.ToString());
                }
            }
        }

        private int FindNearestFrame(List<FrameItem> frames, int startPos, uint frameTime)
        {
            int index = 0;
            uint oldDiff = uint.MaxValue;

            for (index = startPos; index < frames.Count; index++)
            {
                FrameItem frame = frames[index];

                if (frame.TimeStamp == 0)
                {
                    continue;
                }

                if (frame.TimeStamp < frameTime) // fps에 딱 맞는 시점보다 먼저 갭처된 프레임의 경우 시간 차만 구해두고,
                {
                    oldDiff = frameTime - frame.TimeStamp;
                    continue;
                }

                uint diff = frame.TimeStamp - frameTime;
                if (oldDiff < diff) // fps의 시점보다 늦은 프레임이 나왔을 때, 앞 뒤 프레임과의 시간 차에서 낮은 값을 기준으로 한 프레임을 선택
                {
                    return index - 1;
                }

                return index;
            }

            return index - 1;
        }

        public Mat GetMat()
        {
            return _pool.Get();
        }

        public void MarkElapsed(uint elapsed)
        {
            if (_lastCaptured.TimeStamp == 0 || _lastCaptured.Mat == null)
            {
                return;
            }

            Mat data = _pool.Get();
            _lastCaptured.Mat.CopyTo(data);

            FrameItem sameImage = new FrameItem { TimeStamp = _lastCaptured.TimeStamp += elapsed - 1, Mat = data };
            _queue.SendAsync(sameImage);
        }

        public void Write(Mat current)
        {
            _lastCaptured = new FrameItem { TimeStamp = timeGetTime(), Mat = current };
            _queue.SendAsync(_lastCaptured);
        }

        public void Dispose()
        {
            timeEndPeriod(1);

            _queue.SendAsync(new FrameItem());
            _processTaskExit.WaitOne(1000);

            (_pool as IDisposable).Dispose();
        }
    }
}
